class DataLoader:
    # Simple dataloader placeholder for the training demo.
    def __init__(self, batch_size=10, shuffle_batch=True):
        self.summary = f"DataLoader:\n-{batch_size=}\n-{shuffle_batch=}\n"