"""
constants/DSim.py

    AI Hydra
    Author: Nadim-Daniel Ghaznavi
    Copyright: (c) 2025 Nadim-Daniel Ghaznavi
    GitHub: https://github.com/NadimGhaznavi/ai-hydra
    Website: https://aihydra.osoyalce.com
    License: GPL 3.0
"""


class DSim:
    BOARD_SIZE: int = 20
